
#include <cstdlib>
#include <cassert>
#include <iostream>
#include <string>
#include "rb_tree.hpp"


using std::cout;
using std::endl;


template<typename Key>
rb_tree<Key>::rb_tree() : _root()
{
}


template<typename Key>
rb_tree<Key>::~rb_tree()
{
	cout << "rb_tree::~rb_tree():" << endl;
	destroy_subtree(_root);
}


template<typename Key>
shared_ptr<node<Key> > rb_tree<Key>::find(Key key)
{
	if (_root.use_count() == 0)
		return shared_ptr<node<Key> >();
	
	shared_ptr<node<Key> > current(_root);
	while (!current->is_sentinel())
		if (key < current->key)
			current = current->left;
		else if (key > current->key)
			current = current->right;
		else
			return current;

	return shared_ptr<node<Key> >();
}


template<typename Key>
void rb_tree<Key>::insert(Key key)
{
	shared_ptr<node<Key> > new_node = node<Key>::create_node(key);
	
	if (_root.get() == 0)
	{
		_root = new_node;
		_root->color = node<Key>::BLACK;
		return;
	}
	
	// find the right place to insert
	shared_ptr<node<Key> > current = _root;
	while (true)
	{
		if (key < current->key)
		{
			if (current->left->is_sentinel())
			{
				current->left = new_node;
				new_node->parent = current;
				break;
			}
			current = current->left;
		}
		else
		{
			if (current->right->is_sentinel())
			{
				current->right = new_node;
				new_node->parent = current;
				break;
			}
			current = current->right;
		}
	}
	
	if (new_node->parent == _root)
		return;
	
	// fix red-black properties
	assert(new_node->parent->parent.use_count() > 0);
	while (new_node->parent.use_count() > 0 && new_node->parent->color == node<Key>::RED)
	{
		assert(new_node->parent->parent.use_count() > 0);
			
		// cases L1-L3
		if (new_node->parent == new_node->parent->parent->left)
		{
			shared_ptr<node<Key> > uncle = new_node->parent->parent->right;
			if (uncle.use_count() > 0 && uncle->color == node<Key>::RED) // L1
			{
				new_node->parent->color = node<Key>::BLACK;
				uncle->color = node<Key>::BLACK;
				new_node->parent->parent->color = node<Key>::RED;
				new_node = new_node->parent->parent;
			}
			else
			{
				if (new_node == new_node->parent->right) // L2
				{
					rotate_left(new_node->parent);
					new_node = new_node->left;
				}
				// L3
				new_node->parent->color = node<Key>::BLACK;
				new_node->parent->parent->color = node<Key>::RED;
				rotate_right(new_node->parent->parent);
			}
		}
		// cases R1-R3
		else if (new_node->parent == new_node->parent->parent->right)
		{
			shared_ptr<node<Key> > uncle = new_node->parent->parent->left;
			if (uncle.use_count() > 0 && uncle->color == node<Key>::RED) // R1
			{
				new_node->parent->color = node<Key>::BLACK;
				uncle->color = node<Key>::BLACK;
				new_node->parent->parent->color = node<Key>::RED;
				new_node = new_node->parent->parent;
			}
			else
			{
				if (new_node == new_node->parent->left) // R2
				{
					rotate_right(new_node->parent);
					new_node = new_node->right;
				}
				// R3
				new_node->parent->color = node<Key>::BLACK;
				new_node->parent->parent->color = node<Key>::RED;
				rotate_left(new_node->parent->parent);
			}
		}
		else
			assert(false);
	}
	_root->color = node<Key>::BLACK;
}


template<typename Key>
void rb_tree<Key>::remove(Key key)
{
	if (_root.use_count() == 0)
		return;

	shared_ptr<node<Key> > del_node = find(key);
	if (del_node.use_count() == 0)
		return;

	if (del_node == _root && _root.use_count() == 1)
	{
		_root.reset();
		return;
	}

	// red leaf is deleted only
	if (del_node->color == node<Key>::RED && del_node->left->is_sentinel() && del_node->right->is_sentinel())
	{
		if (del_node == del_node->parent->left)
			del_node->parent->left = node<Key>::create_sentinel(del_node->parent);
		else if (del_node == del_node->parent->right)
			del_node->parent->right = node<Key>::create_sentinel(del_node->parent);
		else
			assert(false);
		return;
	}

	shared_ptr<node<Key> > succ(successor(del_node));
	shared_ptr<node<Key> > pred(predecessor(del_node));
	shared_ptr<node<Key> > succ_pred;
	shared_ptr<node<Key> > extra_black;

	// for black leaf, take it's parent as predecessor/successor
	if (succ.get() == 0 && pred.get() == 0)
	{
		shared_ptr<node<Key> > succ_pred_p = del_node->parent;
		succ_pred = succ_pred_p;
		if (del_node == succ_pred_p->left)
		{
			succ_pred_p->left = node<Key>::create_sentinel(succ_pred_p);
			extra_black = succ_pred_p->left;
		}
		else if (del_node == succ_pred_p->right)
		{
			succ_pred_p->right = node<Key>::create_sentinel(succ_pred_p);
			extra_black = succ_pred_p->right;
		}
		else
			assert(false);
	}
	else if (succ.get() != 0)
	{
		shared_ptr<node<Key> > succ_pred_p = succ->parent;
		succ_pred = succ;
		del_node->key = succ->key;
		// reconnect succ_pred_p with succ_pred's right child
		if (succ == succ_pred_p->left)
		{
			if (succ->right.get() != 0)
			{
				succ_pred_p->left = succ->right;
				succ->right->parent = succ_pred_p;
			}
			else
				succ_pred_p->left = node<Key>::create_sentinel(succ_pred_p);
			extra_black = succ_pred_p->left;
		}
		else if (succ == succ_pred_p->right)
		{
			succ_pred_p->right = node<Key>::create_sentinel(succ_pred_p);
			extra_black = succ_pred_p->right;
		}
	}
	else if (pred.get() != 0)
	{
		shared_ptr<node<Key> > succ_pred_p = pred->parent;
		succ_pred = pred;
		del_node->key = pred->key;
		// reconnect succ_pred_p with succ_pred's left child
		if (pred == succ_pred_p->right)
		{
			if (pred->left.get() != 0)
			{
				succ_pred_p->right = pred->left;
				pred->left->parent = succ_pred_p;
			}
			else
				succ_pred_p->right = node<Key>::create_sentinel(succ_pred_p);
			extra_black = succ_pred_p->right;
		}
		else if (pred == succ_pred_p->left)
		{
			succ_pred_p->left = node<Key>::create_sentinel(succ_pred_p);
			extra_black = succ_pred_p->left;
		}
	}

	if (succ_pred->color == node<Key>::BLACK || (extra_black->parent == succ_pred))
	{
		// fix red-black properties by placing extra blackness of succ_pred
		while (extra_black != _root && extra_black->color == node<Key>::BLACK)
		{
			if (extra_black == extra_black->parent->left) // cases L1-L4
			{
				shared_ptr<node<Key> > sibling = extra_black->parent->right;
				if (sibling->color == node<Key>::RED) // L1
				{
					sibling->color = node<Key>::BLACK;
					extra_black->parent->color = node<Key>::RED;
					rotate_left(extra_black->parent);
					sibling = extra_black->parent->right;
				}
				if (!sibling->is_sentinel() && sibling->color == node<Key>::BLACK && sibling->left->color == node<Key>::BLACK && sibling->right->color == node<Key>::BLACK) // L2
				{
					sibling->color = node<Key>::RED;
					extra_black = extra_black->parent;
					if (extra_black->color == node<Key>::RED)
						break;
				}
				else
				{
					assert(!sibling->is_sentinel());

					if (sibling->right->color == node<Key>::BLACK) // L3
					{
						sibling->left->color = node<Key>::BLACK;
						sibling->color = node<Key>::RED;
						rotate_right(sibling);
						sibling = extra_black->parent->right;
					}
					// L4
					assert(sibling->color == node<Key>::BLACK);
					sibling->color = extra_black->parent->color;
					extra_black->parent->color = node<Key>::BLACK;
					sibling->right->color = node<Key>::BLACK;
					rotate_left(extra_black->parent);
					extra_black = _root; // break the loop
				}
			}
			else if (extra_black == extra_black->parent->right) // cases R1-R4
			{
				shared_ptr<node<Key> > sibling = extra_black->parent->left;
				if (sibling->color == node<Key>::RED) // R1
				{
					sibling->color = node<Key>::BLACK;
					extra_black->parent->color = node<Key>::RED;
					rotate_right(extra_black->parent);
					sibling = extra_black->parent->left;
				}
				if (!sibling->is_sentinel() && sibling->color == node<Key>::BLACK && sibling->right->color == node<Key>::BLACK && sibling->left->color == node<Key>::BLACK) // R2
				{
					sibling->color = node<Key>::RED;
					extra_black = extra_black->parent;
					if (extra_black->color == node<Key>::RED)
						break;
				}
				else
				{
					assert(!sibling->is_sentinel());

					if (sibling->left->color == node<Key>::BLACK) // R3
					{
						sibling->right->color = node<Key>::BLACK;
						sibling->color = node<Key>::RED;
						rotate_left(sibling);
						sibling = extra_black->parent->left;
					}
					// R4
					assert(sibling->color == node<Key>::BLACK);
					sibling->color = extra_black->parent->color;
					extra_black->parent->color = node<Key>::BLACK;
					sibling->left->color = node<Key>::BLACK;
					rotate_right(extra_black->parent);
					extra_black = _root; // break the loop
				}
			}
			else
				assert(false);
		}
		extra_black->color = node<Key>::BLACK;
	}
}


template<typename Key>
void rb_tree<Key>::rotate_left(shared_ptr<node<Key> > a_node)
{
	assert(a_node->right != 0);

	shared_ptr<node<Key> > node_p = a_node->parent;
	shared_ptr<node<Key> > node_r = a_node->right;
	shared_ptr<node<Key> > node_rl = a_node->right->left;
	
	node_r->parent = node_p;
	a_node->parent = node_r;
	if (!node_rl->is_sentinel())
		node_rl->parent = a_node;
	a_node->right = node_rl;
	node_r->left = a_node;
	if (node_p.get() != 0)
	{
		if (a_node == node_p->left)
			node_p->left = node_r;
		else if (a_node == node_p->right)
			node_p->right = node_r;
		else
			assert(false);
	}
	
	if (_root == a_node)
		_root = a_node->parent;
}


template<typename Key>
void rb_tree<Key>::rotate_right(shared_ptr<node<Key> > a_node)
{
	assert(a_node->left != 0);

	shared_ptr<node<Key> > node_p = a_node->parent;
	shared_ptr<node<Key> > node_l = a_node->left;
	shared_ptr<node<Key> > node_lr = a_node->left->right;
	
	node_l->parent = node_p;
	a_node->parent = node_l;
	if (!node_lr->is_sentinel())
		node_lr->parent = a_node;
	a_node->left = node_lr;
	node_l->right = a_node;
	if (node_p.get() != 0)
	{
		if (a_node == node_p->left)
			node_p->left = node_l;
		else if (a_node == node_p->right)
			node_p->right = node_l;
		else
			assert(false);
	}
	
	if (_root == a_node)
		_root = a_node->parent;
}


template<typename Key>
void rb_tree<Key>::destroy_subtree(shared_ptr<node<Key> > root)
{
	if (root->left.get() != 0)
	{
		root->left->parent.reset();
		destroy_subtree(root->left);
		root->left.reset();
	}

	if (root->right.get() != 0)
	{
		root->right->parent.reset();
		destroy_subtree(root->right);
		root->right.reset();
	}

	root.reset();
}


template<typename Key>
shared_ptr<node<Key> > rb_tree<Key>::predecessor(shared_ptr<node<Key> > a_node)
{
	if (a_node->is_sentinel() || _root.get() == 0 || a_node->left->is_sentinel())
		return shared_ptr<node<Key> >();
	
	shared_ptr<node<Key> > pred = a_node->left;
	while (!pred->right->is_sentinel())
		pred = pred->right;
	return pred;
}


template<typename Key>
shared_ptr<node<Key> > rb_tree<Key>::successor(shared_ptr<node<Key> > a_node)
{
	if (a_node->is_sentinel() || _root.get() == 0 || a_node->right->is_sentinel())
		return shared_ptr<node<Key> >();
	
	shared_ptr<node<Key> > succ = a_node->right;
	while (!succ->left->is_sentinel())
		succ = succ->left;
	return succ;
}


int main()
{
	rb_tree<int> tree;
	
	tree.insert(11);
	tree.insert(7);
	tree.insert(5);
	tree.insert(2);
	tree.insert(1);
	tree.insert(22);
	tree.insert(14);
	tree.insert(12);
	tree.insert(25);
	tree.insert(3);
	tree.insert(4);
	tree.insert(18);
	tree.insert(24);
	tree.insert(26);
	tree.insert(29);
	tree.insert(31);

	tree.remove(18);
	tree.remove(3);
	tree.remove(7);

	return EXIT_SUCCESS;
}
 